# MODEL PRICES
model_prices = {
    "gpt-3.5-turbo": {
        "prompt_tokens": 0.001 / 1000,
        "completion_tokens": 0.002 / 1000,
    },
    "gpt-4-1106-preview": {
        "prompt_tokens": 0.01 / 1000,
        "completion_tokens": 0.03 / 1000,
    },
    "llama2-70b-chat": {
        "prompt_tokens": 0.03 / 1000,
        "completion_tokens": 0.03 / 1000,
    },
}
